from typing import Any, Dict, List, Optional

from pydantic import BaseModel

from .base_models import SubmissionType
from .gql_client import HLClient

__all__ = ["update_task_result"]


class UpdateTaskResultPayload(BaseModel):
    submission: SubmissionType
    errors: List[Any]


def update_task_result(
    client: HLClient,
    task_id: str,
    status: str,
    observations: List[Dict] = [],
    background_info_layer_file_data: Optional[Dict] = None,
):
    assert status in ("PENDING", "RUNNING", "FAILED", "SUCCESS")

    kwargs = {
        "id": task_id,
        "status": status,
        "eavtAttributes": observations,
    }
    if background_info_layer_file_data is not None:
        kwargs["backgroundInfoLayerFileData"] = background_info_layer_file_data

    result = client.updateTaskResultV2(return_type=UpdateTaskResultPayload, **kwargs)
    return result



import json
import sys
from collections import defaultdict
from pathlib import Path
from typing import Dict, Optional, Tuple

from .base_capability import Capability, ContextPipelineElement, Entities, StreamEvent

__all__ = ["EntityWrite"]


def serialize_entity(entity):
    # ToDo: Make this more general
    if entity.annotations:
        annotations = [
            {
                "location": (
                    entity.annotations[0].location.wkt if entity.annotations[0].location is not None else None
                ),
                "observations": [o.model_dump() for o in entity.annotations[0].observations],
            }
        ]
    else:
        annotations = []

    if entity.global_observations:
        global_observations = [o.model_dump() for o in entity.global_observations]
    else:
        global_observations = []

    result = {
        "id": str(entity.id),
        "annotations": annotations,
        "global_observations": global_observations,
    }
    return result


class EntityWrite(Capability):
    """
    {
      "name": "EntityWrite",
      "parameters": {
        "output_per_frame": false,
        "output_file_path": ""
      },
      "input": [
        {
          "name": "entities",
          "type": "dict"
        }
      ],
      "output": [],
      "deploy": {
        "local": {
          "module": "highlighter.agent.capabilities"
        }
      }
    }
    """

    class DefaultStreamParameters(Capability.DefaultStreamParameters):
        # if True: write entities each time process_frame is called
        # if False: compile entities and write all to a singel file when stop_stream is called
        output_per_frame: bool = False

        # can use substitution for output_file_path
        # if output_per_frame is True:
        #   - frame_id
        #   - source_file_name -> equivalent to Path(source_file_location).stem
        #
        # if output_per_frame is False:
        #   - source_file_name -> equivalent to Path(source_file_location).stem
        #
        # If slashes are use the directories will be created.
        #  eg: output_file_path = "my_output/frame_{frame_id}_{source_file_name}.json
        output_file_path: str = ""

    @property
    def output_per_frame(self) -> bool:
        value, _ = self._get_parameter("output_per_frame")
        return value

    @property
    def output_file_path(self) -> str:
        value, _ = self._get_parameter("output_file_path")
        return value

    @classmethod
    def get_capability_input_output_deploy(cls):
        return {
            "input": [{"name": "entities", "type": "Entities"}],
            "output": [{}],
            "deploy": {"local": {"module": "highlighter.agent.capabilities"}},
        }

    def __init__(self, context: ContextPipelineElement):
        super().__init__(context)
        self.entities = defaultdict(list)

    def start_stream(self, stream, stream_id) -> Tuple[StreamEvent, Optional[str]]:
        return StreamEvent.OKAY, None

    def _get_source_file_location(self, stream):
        # ToDo: Find a better palce to put/get this from
        #       see also, ImageDataSource.process_frame
        source_info = stream.get("source_info", {})
        source_file_location = source_info.get("source_file_location", None)
        if source_file_location is not None:
            return Path(source_file_location)
        return None

    def process_frame(self, stream, entities) -> Tuple[StreamEvent, Dict]:

        source_file_location = self._get_source_file_location(stream)

        serializable_entities = {
            "frame_id": stream["frame_id"],
            "source_file_location": str(source_file_location),
            **self._dump_frame_entities(entities),
        }
        if self.output_per_frame:
            # Dump to std out
            json_entites = json.dumps(serializable_entities, indent=2, sort_keys=True)
            print(json_entites, file=sys.stdout)

            if self.output_file_path:
                # Dump to file
                frame_id = stream["frame_id"]
                output_file_path = self.output_file_path.format(
                    frame_id=frame_id,
                    source_file_name=source_file_location.stem if source_file_location is not None else "",
                )
                Path(output_file_path).parent.mkdir(exist_ok=True, parents=True)
                with open(output_file_path, "w") as f:
                    f.write(json_entites)
        else:
            self.entities[stream["stream_id"]].append(serializable_entities)

        return StreamEvent.OKAY, {}

    def _dump_frame_entities(self, frame_entities: Dict):
        serializable_entities = {}
        for ent_id, entity in frame_entities.items():
            serializable_entities[str(ent_id)] = serialize_entity(entity)
        return serializable_entities

    def stop_stream(self, stream, stream_id) -> Tuple[StreamEvent, Optional[str]]:
        entities_json = json.dumps(self.entities[stream_id], indent=2, sort_keys=True)
        if not self.output_per_frame:

            if self.output_file_path:
                source_file_location = self._get_source_file_location(stream)
                output_file_path = self.output_file_path.format(
                    source_file_name=source_file_location.stem if source_file_location is not None else "",
                )
                with open(output_file_path, "w") as f:
                    f.write(entities_json)

            else:
                print(entities_json, file=sys.stdout)
        return StreamEvent.OKAY, None



from typing import List, Sequence, Tuple, Union

import numpy as np
from shapely import affinity
from shapely import geometry as geom
from shapely import make_valid
from shapely.ops import unary_union


def polygon_from_left_top_width_height_coords(coords: Union[Tuple[float, float, float, float], np.ndarray]):
    """Create a LocationAttributeValue for a box in the form [x,y,w,h]

    Args:
        coords: A sequence of box in the form [x,y,w,h]
    """
    x0, y0, w, h = coords
    x1 = x0 + w
    y1 = y0 + h
    _coords = [[x0, y0], [x1, y0], [x1, y1], [x0, y1], [x0, y0]]
    return geom.Polygon(_coords)


def try_make_polygon_valid_if_invalid(shape):
    if shape.is_valid:
        return shape

    valid_shape = make_valid(shape)
    if not isinstance(valid_shape, (geom.MultiPolygon, geom.Polygon)):
        raise ValueError(f"Invalid Polygon/MultiPolygon {shape}")
    return valid_shape


def polygon_from_coords(
    coords: Sequence[
        Union[Tuple[float, float], Tuple[float, float, float], np.ndarray, Sequence[geom.Point]]
    ],
    fix_invalid_polygons: bool = False,
):
    """
    Args:
        coords: A sequence of (x, y [,z]) numeric coordinate pairs or triples, or
        an array-like with shape (N, 2) or (N, 3).
        Also can be a sequence of Point objects.
    """
    if len(coords) < 3:
        raise ValueError(f"Polygon must have at least 3 coordinates: {coords}")

    value = geom.Polygon(coords)
    if fix_invalid_polygons:
        value = try_make_polygon_valid_if_invalid(value)
    return value


def polygon_from_tlbr(x: Tuple[int, int, int, int]) -> geom.Polygon:
    """
    from top left bottom right format
    """
    top_left = x[0], x[1]
    top_right = x[2], x[1]
    bottom_right = x[2], x[3]
    bottom_left = x[0], x[3]
    return geom.Polygon(np.array([top_left, top_right, bottom_right, bottom_left]))


def multipolygon_from_coords(
    coords: Sequence[Sequence[Tuple[float, float]]],
    fix_invalid_polygons: bool = False,
):
    """
    Args:
        coords: A nested sequence of (x, y) numeric coordinate pairs, or
        an array-like with shape (N, 2).
    """
    shapes = []
    for poly_xys in coords:
        shape = geom.Polygon(poly_xys)
        if fix_invalid_polygons:
            shape = try_make_polygon_valid_if_invalid(shape)
        shapes.append(shape)

    if len(shapes) == 1:
        value = shapes[0]
    else:
        value = unary_union(shapes)
    return value


def get_top_left_bottom_right_coordinates(
    value: geom.Polygon, scale: float = 1.0, scale_about_origin: bool = True, pad: int = 0
) -> Tuple[int, int, int, int]:
    """
    to top left bottom right format
    """
    bounds: geom.Polygon = geom.box(*value.bounds)

    if scale_about_origin:
        bounds = affinity.scale(bounds, xfact=scale, yfact=scale, origin=(0, 0, 0))
    else:
        bounds = affinity.scale(bounds, xfact=scale, yfact=scale, origin="center")

    bounds = bounds.buffer(pad, join_style="bevel")
    return bounds.bounds



import json
import sys
from enum import Enum
from pathlib import Path
from tempfile import mkdtemp
from typing import List
from uuid import UUID

import click

from ..agent.capabilities import sources as source_capabilities
from ..agent.capabilities import targets as target_capabilities
from ..agent.capabilities.base_capability import Capability


def _is_uuid(s):
    try:
        _ = UUID(s)
        return True
    except Exception as _:
        return False


@click.group("agent")
@click.pass_context
def agent_group(ctx):
    pass


DATASOURCE_TAGS = ("--data-source", "-ds")
DATATARGET_TAGS = ("--data-target", "-dt")
STREAM_PARAMS_TAGS = ("--stream-parameters", "-sp")


class AgentRunHelpFormatter(click.HelpFormatter):
    def __init__(self, *args, **kwargs):
        (self._capability_help, self._capability_cli_args) = self._get_capability_help()
        super().__init__(*args, **kwargs)

    def _get_capability_help(self):
        argv = sys.argv
        capability_name = argv[-2]
        capability_cls: Capability
        if (capability_cls := getattr(source_capabilities, capability_name, None)) is not None:
            capability_def = capability_cls.get_capability_definition()
            capability_tag = " ".join(DATASOURCE_TAGS)
        elif (capability_cls := getattr(target_capabilities, capability_name, None)) is not None:
            capability_def = capability_cls.get_capability_definition()
            capability_tag = " ".join(DATATARGET_TAGS)
        else:
            capability_def = None
            capability_tag = ""

        if capability_def is None:
            return None, None

        help_str = """\b
__CAPABILITY_NAME__

\b
Params:
__PARAMS__
    
\b
Inputs:
__INPUTS__
    
\b
Outputs:
__OUTPUTS__
            """
        params_items = ["\b"]
        params_args_str = [capability_tag, capability_name]
        help_str = help_str.replace("__CAPABILITY_NAME__", capability_def["name"])

        for field_name, default_value in capability_cls.default_stream_parameters().model_dump().items():
            field_type = type(default_value)

            if issubclass(field_type, Enum):
                choices = "|".join([member.name for member in default_value.__class__])
                params_items.append(f"  {field_name}: {choices}")
                params_args_str.append(f"{STREAM_PARAMS_TAGS[1]} {field_name}={default_value.value}")
            else:
                params_items.append(f"  {field_name}: {field_type.__name__}")
                params_args_str.append(f"{STREAM_PARAMS_TAGS[1]} {field_name}={default_value}")

        help_str = help_str.replace("__PARAMS__", "\n".join(params_items))
        params_args_str.append("...")

        input_items = ["\b"]
        for i in capability_def["input"]:
            input_items.append(f"  {i['name']}: {i['type']}")
        help_str = help_str.replace("__INPUTS__", "\n".join(input_items))

        output_items = ["\b"]
        for i in capability_def["output"]:
            output_items.append(f"  {i['name']}: {i['type']}")
        help_str = help_str.replace("__OUTPUTS__", "\n".join(output_items))

        return click.wrap_text(help_str, width=200), " ".join(params_args_str)

    def write_text(self, text):
        if self._capability_help is None:
            super().write_text(text)
        else:
            super().write_text(self._capability_help)

    def write_heading(self, heading):
        if self._capability_help is None:
            super().write_heading(heading)
        else:
            super().write_heading("")

    def write_usage(self, prog, args="", prefix="Usage: "):
        if self._capability_help is None:
            super().write_usage(prog, args, prefix=prefix)
        else:
            super().write_usage(prog, self._capability_cli_args, prefix="Capability Usage: ")

    def write_dl(
        self,
        rows,
        col_max: int = 30,
        col_spacing: int = 2,
    ) -> None:
        if self._capability_help is None:
            super().write_dl(rows, col_max=col_max, col_spacing=col_spacing)
        else:
            return None


class AgentRunCommand(click.Command):
    def get_help(self, ctx):
        formatter = AgentRunHelpFormatter()
        self.format_help(ctx, formatter)
        return formatter.getvalue()


def parse_stream_params(agent, agent_definition, str_params: List[str]) -> dict:
    # If different Capabilities have clashing parameter names then
    # the parameter in the Capability later in the pipeline will take
    # precedence. To get around this you can prepend the stream parameter
    # name with the Capability name in the cli
    # "-sp CapabilityA.foo=42 -sp CapabilityB.foo=3.14
    stream_params = {}
    for node in agent.pipeline_graph.nodes():
        node_name = node.element.__class__.__name__

        # Start with in code params
        default_stream_parameters = node.element.default_stream_parameters().model_dump()

        # Overwite with global pipeline definition params
        global_pipeline_definition_params = {
            k: v for k, v in agent_definition.parameters if k in default_stream_parameters
        }
        default_stream_parameters.update(global_pipeline_definition_params)

        # Overwite with per element pipeline definition paras
        element_definition = [e for e in agent_definition.elements if e.name == node_name][0]
        pipeline_element_definition_params = {
            k: v for k, v in element_definition.parameters.items() if k in default_stream_parameters
        }
        default_stream_parameters.update(pipeline_element_definition_params)

        # Overwite with cli params
        cli_params = {}
        for param_str in str_params:
            k, v = param_str.split("=")
            cli_params[k] = v

        for param_name in default_stream_parameters:

            # preference fully qualified param names
            node_param_name = f"{node_name}.{param_name}"
            if node_param_name in cli_params:
                override_value = cli_params[node_param_name]
                default_stream_parameters[param_name] = override_value

            # fall back to unqualified param names
            elif param_name in cli_params:
                override_value = cli_params[param_name]
                default_stream_parameters[param_name] = override_value

        ele_stream_params = node.element.DefaultStreamParameters(**default_stream_parameters).model_dump()
        stream_params.update({f"{node_name}.{k}": v for k, v in ele_stream_params.items()})

    return stream_params


@agent_group.command("run", cls=AgentRunCommand)
@click.option(*DATASOURCE_TAGS, type=click.Choice(source_capabilities.__all__), default=None)
@click.option(*DATATARGET_TAGS, type=click.Choice(target_capabilities.__all__), default=None)
@click.option("--stream-id", "-s", type=int, default=1)
@click.option(*STREAM_PARAMS_TAGS, type=str, multiple=True, default=[])
@click.option("--dump-definition", type=str, default=None)
@click.argument("agent_definition", type=click.Path(dir_okay=False, exists=False))
@click.argument("input_data", type=click.Path(exists=False), default="-", required=False)
@click.pass_context
def _run(
    ctx,
    data_source,
    data_target,
    stream_id,
    stream_parameters,
    dump_definition,
    agent_definition,
    input_data,
):
    """Run a local Highlighter Agent to process data.

    To feed data into the agent you have 3 options:

    \b 
    1. Add a DataSource Capability to the Agent definition:
        # agent_definition.json has an ImageDataSource Capability
        hl agent run agent_definition.json image.jpg

    \b
    2. Use the cli --data-source flag:
        # Add an ImageDataSource to the front of your Agent
        hl agent run --data-source ImageDataSource agent_definition.json image.jpg

    \b
    3. Pass frame params directly
        # Your Capabilities process_frame functions have named 
        # arguments "foo" and "bar"
        cat frame_data.json | hl agent run agent_definition.json 
        or
        hl agent run agent_definition.json '[{"foo": 123, "bar": 456}]'

    \b
    If --data-source or --data-target are not specified then it is assumed one
    or both of them are defined in the agent_definition.

    \b
    To get specific help for a --data-source or --data-target use --help 
    after specifying it, eg:
        hl agent run --data-source ImageDataSource --help

    \b 
    Examples:
        # Process a single image from path
        hl agent run --data-source ImageDataSource PIPELINE.json image.jpg 

    \b 
        # Process a single image from bytes
        cat image.jpg | hl agent run --data-source ImageDataSource -sp read_image_bytes=True PIPELINE.json

    \b 
        # Process many images from paths
        find ./image/dir -name "*.jpg" | hl agent run --data-source ImageDataSource PIPELINE.json

    \b 
       # Process a video from frame 10 to frame 20
        hl agent run -ds VideoDataSource \\
                -sp start_frame=10 \\
                -sp stop_frame=20 \\
                PIPELINE.json \\
                video.mp4

    """
    from highlighter.agent import init_agent

    if Path(agent_definition).exists():
        with Path(agent_definition).open("r") as f:
            definition_dict = json.load(f)

        elements = definition_dict["elements"]
        graph_def = definition_dict["graph"][0]
        if data_source is not None:
            source_cls: Capability = getattr(source_capabilities, data_source)
            source_definition = source_cls.get_capability_definition()
            elements = [source_definition] + elements
            graph_def = f"({source_definition['name']} " + graph_def[1:]

        if data_target is not None:
            target_cls: Capability = getattr(target_capabilities, data_target)
            target_definition = target_cls.get_capability_definition()
            elements = elements + [target_definition]
            graph_def = graph_def[:-1] + f" {target_definition['name']})"

        definition_dict["elements"] = elements
        definition_dict["graph"] = [graph_def]

        agent_def_json = json.dumps(definition_dict, sort_keys=True, indent=2)

        if dump_definition is not None:
            agent_path = Path(dump_definition)
        else:
            agent_path = Path(mkdtemp()) / "agent_def.json"

        with agent_path.open("w") as f:
            f.write(agent_def_json)

        name = Path(agent_definition).name

    elif _is_uuid(agent_definition):
        raise NotImplementedError("Need to implement pulling the agent_definition from Highlighter")
    else:
        if not Path(agent_definition).suffix == ".json":
            raise SystemExit(f"agent_definition '{agent_definition}' path does not exist")
        else:
            raise SystemExit(f"agent_definition '{agent_definition}' id does not exist")

    agent, agent_definition = init_agent(agent_path, name)

    _stream_parameters: dict = parse_stream_params(agent, agent_definition, stream_parameters)

    def agent_driven_by_datasource(agent):
        head_capability_name = [x for x in agent.pipeline_graph][0].name
        if getattr(source_capabilities, head_capability_name, None) is not None:
            return True, head_capability_name
        return False, None

    (driven_by_datasource, data_source_capability_name) = agent_driven_by_datasource(agent)

    if driven_by_datasource:
        if input_data == "-":
            # if "-" is passed as input_data read bytes from stdin
            if not sys.stdin.isatty():
                _stream_parameters[f"{data_source_capability_name}.source_inputs"] = sys.stdin.buffer.read()
            else:
                # Assume source_inputs are in the agent definition
                pass

        else:
            # The agent source_inputs is the path to the file.
            # It is expected the data-source capability knows how to handel that.
            _stream_parameters[f"{data_source_capability_name}.source_inputs"] = [input_data]

        agent.create_stream(stream_id, _stream_parameters)
        agent.run(mqtt_connection_required=False)

    else:
        if input_data == "-":
            # if "-" is passed as input_data read bytes from stdin
            if not sys.stdin.isatty():
                frame_data = json.load(sys.stdin)
            else:
                frame_data = [{}]

        else:
            # The agent source_inputs is the path to the file.
            # It is expected the data-source capability knows how to handel that.
            frame_data = json.loads(input_data)

        if isinstance(frame_data, dict):
            frame_data = [frame_data]

        agent.create_stream(stream_id, _stream_parameters)
        stream = {"stream_id": 1}
        for frame_id, frame in enumerate(frame_data):
            stream["frame_id"] = frame_id
            agent.process_frame(stream, frame)



from datetime import datetime
from json import JSONEncoder
from uuid import UUID

__all__ = ["HLJSONEncoder"]


class HLJSONEncoder(JSONEncoder):
    def default(self, obj):
        # ToDo: Maybe add more custom endoders for various
        # HLBaseModels
        if isinstance(obj, UUID):
            return str(obj)
        elif isinstance(obj, datetime):
            return obj.isoformat()
        return super().default(obj)



import os

import aiko_services as aiko

__all__ = [
    "init_agent",
    "run_agent",
]


def init_agent(definition_pathname, name):

    if not os.path.exists(definition_pathname):
        raise SystemExit(f"Error: PipelineDefinition not found: {definition_pathname}")

    pipeline_definition = aiko.PipelineImpl.parse_pipeline_definition(definition_pathname)
    name = name if name else pipeline_definition.name

    init_args = aiko.pipeline_args(
        name,
        protocol=aiko.PROTOCOL_PIPELINE,
        definition=pipeline_definition,
        definition_pathname=definition_pathname,
    )
    pipeline = aiko.compose_instance(aiko.PipelineImpl, init_args)
    return pipeline, pipeline_definition


def run_agent(pipeline, stream_id, stream_parameters):

    pipeline.create_stream(stream_id, dict(stream_parameters))



from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
from PIL import Image
from pydantic import BaseModel

from highlighter.client import HLClient, download_bytes, get_presigned_url
from highlighter.client.io import _pil_open_image_bytes, _pil_open_image_path

from .base_capability import Capability, StreamEvent

__all__ = [
    "ImageDataSource",
    "TextDataSource",
]


class TextDataSource(Capability):
    """

    Example:
        # process a single string
        hl agent run --data-source TextDataSource PIPELINE.json "tell me a joke."

        # process many text files
        ToDo

        # Read from stdin
        cat file | hl agent run --data-source TextDataSource -sp read_stdin=true PIPELINE.json
    """

    class DefaultStreamParameters(Capability.DefaultStreamParameters):
        source_inputs: List = []
        read_stdin: bool = False

    @property
    def read_stdin(self) -> bool:
        value, _ = self._get_parameter("read_stdin")
        return value

    @property
    def source_inputs(self) -> bool:
        value, _ = self._get_parameter("source_inputs")
        return value

    @classmethod
    def get_capability_input_output_deploy(cls) -> Dict:
        return {
            "input": [{"name": "text", "type": "str"}],
            "output": [{"name": "text", "type": "str"}],
            "deploy": {"local": {"module": "highlighter.agent.capabilities"}},
        }

    def _decode_byte_string_to_locations(self, byte_data) -> List[str]:
        separator = b"\x1c"  # ASCII 28 (File Separator)
        byte_lines = byte_data.split(separator)
        text_lines = []
        for byte_string in byte_lines:
            try:
                text_string = byte_string.decode("utf-8")
                text_lines.append(text_string)
            except UnicodeDecodeError:
                raise ValueError("The byte string could not be decoded using UTF-8 encoding.")

        return text_lines

    def get_frame_generator(self, stream, text_lines):
        class FrameGenerator:
            def __init__(self, text_lines, logger):
                self.text_lines = text_lines
                self._logger = logger
                self._logger.info(f"Init FrameGenerator with {len(self.text_lines)}")

            def __call__(self, stream):
                if len(self.text_lines):
                    text = self.text_lines.pop(0)
                    frame_id = stream["frame_id"]
                    self._logger.info(f"generating frame: {frame_id} - {text}")
                    return StreamEvent.OKAY, {"text": text}
                return StreamEvent.STOP, None

        return FrameGenerator(text_lines, self.logger)

    def start_stream(self, stream, stream_id) -> Tuple[StreamEvent, Optional[str]]:

        if stream["stream_id"] == 0:  # TODO: "stream_required"
            raise SystemExit("Must create a stream")

        if self.read_stdin:
            if not isinstance(self.source_inputs, bytes):
                raise ValueError(
                    f"If {self.__class__.__name__}.read_stdin is"
                    " True source_inputs must be bytes, got:"
                    f" {self.source_inputs}. Try setting read_stdin=False."
                )
            text_lines = [self.source_inputs]
        elif isinstance(self.source_inputs, list):
            text_lines = self.source_inputs
        elif isinstance(self.source_inputs, bytes):
            text_lines = self._decode_byte_string_to_locations(self.source_inputs)
        else:
            return (
                StreamEvent.ERROR,
                f"Invalid value for source_inputs, got: {self.source_inputs}",
            )
        frame_generator = self.get_frame_generator(stream, text_lines)

        self.create_frames(stream, frame_generator, rate=1)
        return StreamEvent.OKAY, None

    def process_frame(self, stream, text: str) -> Tuple[StreamEvent, Dict]:
        return StreamEvent.OKAY, {"text": text}

    def stop_stream(self, stream, stream_id) -> Tuple[StreamEvent, Optional[str]]:
        stream["terminate"] = True
        self.stop()
        return StreamEvent.OKAY, None


class OutputType(str, Enum):
    numpy = "numpy"
    pillow = "pillow"


class ImageDataSource(Capability):
    """

    Example:
        # process a single image
        hl agent run PIPELINE.json --data-source ImageDataSource image.jpg

        # process many images
        find image/dir/ -n "*.jpg" | hl agent run PIPELINE.json --data-source ImageDataSource
    """

    class DefaultStreamParameters(BaseModel):
        source_inputs: List = []
        read_image_bytes: bool = False
        output_type: OutputType = OutputType.numpy

    @property
    def read_image_bytes(self) -> bool:
        value, _ = self._get_parameter("read_image_bytes")
        return value

    @property
    def output_type(self) -> OutputType:
        value, _ = self._get_parameter("output_type")
        return value

    @property
    def source_inputs(self) -> bool:
        value, _ = self._get_parameter("source_inputs")
        return value

    @classmethod
    def get_capability_input_output_deploy(cls) -> Dict:
        return {
            "input": [{"name": "image", "type": "Union[str, Path, bytes]"}],
            "output": [{"name": "image", "type": "Union[PIL.Image.Image, np.ndarray]"}],
            "deploy": {"local": {"module": "highlighter.agent.capabilities"}},
        }

    def _decode_byte_string(self, byte_string) -> List[str]:
        try:
            text_string = byte_string.decode("utf-8")
        except UnicodeDecodeError:
            raise ValueError("The byte string could not be decoded using UTF-8 encoding.")

        path_strings = text_string.strip().split("\n")

        return path_strings

    def get_frame_generator(self, stream, images):
        class FrameGenerator:
            def __init__(self, images, logger):
                self.images = images
                self._logger = logger
                self._logger.info(f"Init FrameGenerator with {len(self.images)}")

            def __call__(self, stream):
                if len(self.images):
                    image = self.images.pop(0)
                    frame_id = stream["frame_id"]
                    self._logger.info(f"generating frame: {frame_id}")
                    return StreamEvent.OKAY, {"image": image}
                return StreamEvent.STOP, None

        return FrameGenerator(images, self.logger)

    def start_stream(self, stream, stream_id) -> Tuple[StreamEvent, Optional[str]]:
        super().start_stream(stream, stream_id)

        if stream["stream_id"] == 0:  # TODO: "stream_required"
            raise SystemExit("Must create a stream")

        if self.read_image_bytes:
            if not isinstance(self.source_inputs, bytes):
                raise ValueError(
                    f"If {self.__class__.__name__}.read_image_bytes is"
                    " True source_inputs must be bytes, got:"
                    f" {self.source_inputs}. Try setting read_image_bytes=False."
                )
            images = [self.source_inputs]
        elif isinstance(self.source_inputs, list):
            images = self.source_inputs
        elif isinstance(self.source_inputs, bytes):
            images = self._decode_byte_string(self.source_inputs)
        else:
            return (
                StreamEvent.ERROR,
                f"Invalid value for source_inputs, got: {self.source_inputs}",
            )
        frame_generator = self.get_frame_generator(stream, images)
        self.create_frames(stream, frame_generator, rate=1)
        return StreamEvent.OKAY, None

    def process_frame(self, stream, image: Union[str, Path, bytes]) -> Tuple[StreamEvent, Dict]:
        # ToDo: Find a better palce to put/get this from
        #       see also, EntityWrite.process_frame.
        #       I found it useful when reviewing outputs
        #       to be able to identify the source image

        stream["source_info"] = {}
        if isinstance(image, (str, Path)):
            stream["source_info"] = {"source_file_location": str(image)}

        image_pil: Image.Image = self._read_image(image)
        if self.output_type == OutputType.numpy:
            output = np.array(image_pil)
            self.logger.info(f"image shape: {output.shape}")
        else:
            output = image_pil
            self.logger.info(f"image shape: {output.size}")
        return StreamEvent.OKAY, {"image": output}

    def stop_stream(self, stream, stream_id) -> Tuple[StreamEvent, Optional[str]]:
        stream["terminate"] = True
        self.stop()
        return StreamEvent.OKAY, None

    def _try_cast_int(self, i) -> Optional[int]:
        try:
            _int = int(i)
        except ValueError as _:
            return None
        return _int

    def _read_image_from_location(self, image_location: str) -> Image.Image:
        if image_location.startswith("http:"):
            # ToDo: Add caching
            image_bytes = download_bytes(image_location)
            assert image_bytes is not None
            image = _pil_open_image_bytes(image_bytes)
        elif Path(image_location).exists():
            image = _pil_open_image_path(image_location)
        elif (image_id := self._try_cast_int(image_location)) is not None:
            presigned_url = get_presigned_url(HLClient.get_client(), image_id)
            image_bytes = download_bytes(presigned_url)
            assert image_bytes is not None
            image = _pil_open_image_bytes(image_bytes)
        else:
            raise ValueError(f"Unable to read image from location: {image_location}")

        return image

    def _read_image(self, image: Union[str, Path, bytes]) -> Image.Image:
        if isinstance(image, (str, Path)):
            return self._read_image_from_location(str(image))
        elif isinstance(image, bytes):
            return _pil_open_image_bytes(image)
        else:
            raise ValueError("Invalid image value, expected str|Path|bytes," f"got: {image}")



try:
    import aiko_services
except ModuleNotFoundError as e:
    raise ModuleNotFoundError(
        "To user Highlighter agents you must install aiko_services. "
        "Use `pip install highlighter-sdk[agent]` or, "
        " `pip install aiko_services` manually."
    )
from .agent import *



from datetime import datetime, timezone
from typing import List
from urllib.parse import urlparse

from ..core import DATA_FILE_ATTRIBUTE_UUID
from ..core import LOG as logger
from ..core import PIXEL_LOCATION_ATTRIBUTE_UUID, HLBaseModel
from .gql_client import HLClient
from .training_config import TrainingConfigType, get_training_config

__all__ = ["create_capability_for_artefact"]


def create_capability_for_artefact(
    client: HLClient, artefact_id: str, training_run_id: int, inference_config: dict
):
    # get capability schema via GraphQL
    training_config: TrainingConfigType = get_training_config(client, training_run_id)

    class TrainingRun(HLBaseModel):
        name: str

    training_run_name = client.trainingRun(return_type=TrainingRun, id=training_run_id).name

    class CurrentUser(HLBaseModel):
        id: int

    current_user_id = client.currentUser(return_type=CurrentUser).id

    if "parameters" in inference_config:
        parameters = inference_config["parameters"].copy()
    elif "frame_parameters" in inference_config:
        parameters = inference_config["frame_parameters"].copy()
    else:
        raise KeyError(
            f"'parameters' or 'frame_parameters' cannot be found in inference_config: {inference_config}"
        )

    # pull input filter enum set out of capability schema and put in parameters
    # as this is not currently included in the taxonomy datatypes for inputs and outputs
    if len(training_config.model_schema.model_inputs.filters) > 1:
        raise NotImplementedError("Can't create capabilities from training runs with multiple input filters")
    if len(training_config.model_schema.model_inputs.filters[0]) > 0:
        parameters["enum_array_filter_include_ids"] = (
            training_config.model_schema.get_input_filter_attribute_values()[0]
        )
        parameters["enum_array_filter_attribute"] = (
            training_config.model_schema.get_input_filter_attribute_ids()[0][0]
        )
    # pull confidences out of model schema and put in parameters
    if len(training_config.model_schema.get_head_model_outputs(1)) > 0:
        raise NotImplementedError("Can't create capability from multi-headed model")
    parameters["output_confidence_thresholds"] = {
        str(output.entity_attribute_enum_id): output.conf_thresh
        for output in training_config.model_schema.model_outputs
    }
    machine_agent_version = {
        "frameParameters": parameters,
        "streamParameters": {},
        "deploymentParameters": {},
        "machineAgentTypeId": inference_config["machine_agent_type_id"],
        "trainingRunId": training_run_id,
        "trainingRunArtefactId": artefact_id,
        "title": training_run_name,
        "code": inference_config["code"],
        "codeVersion": "1",
        "publishedById": current_user_id,
        "publishedAt": datetime.now(timezone.utc).isoformat(),
    }

    # create inputs and outputs
    inputs = []
    inputs.append(
        {
            "name": "image",
            "type": "eavt",
            "attribute_id": str(DATA_FILE_ATTRIBUTE_UUID),
            "array_nesting_levels": 0,
        }
    )
    for input in training_config.model_schema.model_inputs.entity_attributes:
        if input.entity_attribute_id == str(PIXEL_LOCATION_ATTRIBUTE_UUID):
            inputs.append(
                {"name": input.entity_attribute_name, "type": "annotation", "array_nesting_levels": 1}
            )
        else:
            inputs.append(
                {
                    "name": input.entity_attribute_name,
                    "type": "eavt",
                    "attribute_id": str(input.entity_attribute_id),
                    "array_nesting_levels": 1,
                }
            )
    if len(training_config.model_schema.model_inputs.filters[0]) > 0:
        inputs.append(
            {
                "name": training_config.model_schema.model_inputs.filters[0][0].entity_attribute_name,
                "type": "eavt",
                "attribute_id": str(
                    training_config.model_schema.model_inputs.filters[0][0].entity_attribute_id
                ),
                "array_nesting_levels": 1,
            }
        )
    heads = []
    for output in training_config.model_schema.model_outputs:
        if output.head not in heads:
            heads.append(output.head)
    outputs = []
    if inference_config["code"] in ["DetectorElement", "Detector"]:
        outputs.append({"name": "pixel_location", "type": "annotation", "array_nesting_levels": 1})
    for head in heads:
        outputs.append(
            {
                "name": training_config.model_schema.get_head_model_outputs(head)[0].entity_attribute_name,
                "type": "eavt",
                "attribute_id": str(
                    training_config.model_schema.get_head_model_outputs(head)[0].entity_attribute_id
                ),
                "array_nesting_levels": 1,
            }
        )

    # run mutations
    class MachineAgentVersion(HLBaseModel):
        id: str

    class CreateMachineAgentVersionPayload(HLBaseModel):
        errors: List[str]
        machine_agent_version: MachineAgentVersion

    create_machine_agent_version_result = client.createMachineAgentVersion(
        return_type=CreateMachineAgentVersionPayload,
        **machine_agent_version,
    )
    if len(create_machine_agent_version_result.errors) > 0:
        raise ValueError(
            f"Could not create machine_agent_version: {create_machine_agent_version_result.errors}"
        )
    machine_agent_version_id = create_machine_agent_version_result.machine_agent_version.id

    class CreatePipelineElementInputPayload(HLBaseModel):
        errors: List[str]

    for i, input in enumerate(inputs):
        if input["type"] == "annotation":
            create_input_result = client.createPipelineElementInput(
                return_type=CreatePipelineElementInputPayload,
                machineAgentVersionId=machine_agent_version_id,
                name=input["name"],
                arrayNestingLevels=input["array_nesting_levels"],
                sortOrder=str(i),
                isAnnotation=True,
            )
        elif input["type"] == "eavt":
            create_input_result = client.createPipelineElementInput(
                return_type=CreatePipelineElementInputPayload,
                machineAgentVersionId=machine_agent_version_id,
                name=input["name"],
                arrayNestingLevels=input["array_nesting_levels"],
                sortOrder=str(i),
                isAnnotation=False,
                entityAttributeId=input["attribute_id"],
            )
        else:
            raise ValueError(f"Input type must be 'annotation' or 'eavt', got '{input['type']}'")
        if len(create_input_result.errors) > 0:
            raise ValueError(f"Could not create agent_element_input: {create_input_result.errors}")

    class CreatePipelineElementOutputPayload(HLBaseModel):
        errors: List[str]

    for i, output in enumerate(outputs):
        if output["type"] == "annotation":
            create_output_result = client.createPipelineElementOutput(
                return_type=CreatePipelineElementOutputPayload,
                machineAgentVersionId=machine_agent_version_id,
                name=output["name"],
                arrayNestingLevels=output["array_nesting_levels"],
                sortOrder=str(i),
                isAnnotation=True,
            )
        elif output["type"] == "eavt":
            create_output_result = client.createPipelineElementOutput(
                return_type=CreatePipelineElementOutputPayload,
                machineAgentVersionId=machine_agent_version_id,
                name=output["name"],
                arrayNestingLevels=output["array_nesting_levels"],
                sortOrder=str(i),
                isAnnotation=False,
                entityAttributeId=output["attribute_id"],
            )
        else:
            raise ValueError(f"Output type must be 'annotation' or 'eavt', got '{output['type']}'")
        if len(create_output_result.errors) > 0:
            raise ValueError(f"Could not create agent_element_output: {create_output_result.errors}")

    logger.info(f"Machine Agent Version {machine_agent_version_id} created and published successfully.")
    endpoint = urlparse(client.endpoint_url)
    endpoint_domain = f"{endpoint.scheme}://{endpoint.netloc}"
    logger.info(
        "You can inspect this machine agent version at "
        f"{endpoint_domain}/admin/machine_agent_versions/{machine_agent_version_id}"
    )

    return machine_agent_version_id



from enum import Enum
from pathlib import Path
from typing import List, Optional, Union
from uuid import UUID
from warnings import warn

import numpy as np
import pandas as pd
import yaml

from highlighter.core import (
    OBJECT_CLASS_ATTRIBUTE_UUID,
    PIXEL_LOCATION_ATTRIBUTE_UUID,
    LabeledUUID,
)
from highlighter.datasets.formats.coco.common import (
    segmentation_to_bbox,
    shapely_to_segmentation,
)
from highlighter.datasets.interfaces import IWriter

PathLike = Union[str, Path]


class YoloWriter(IWriter):
    format_name = "yolo"
    annotation_count = -1
    image_count = -1

    class TASK(Enum):
        DETECT = "detect"
        SEGMENT = "segment"

    def __init__(
        self,
        output_dir: PathLike,
        image_cache_dir: PathLike,
        category_attribute_id: Union[str, UUID, LabeledUUID] = OBJECT_CLASS_ATTRIBUTE_UUID,
        categories: Optional[List[Union[str, UUID]]] = None,
        task: Union[str, TASK] = TASK.DETECT,
    ):
        """Save a Highlighter Dataset object as a YoloV8 Detection dataset

        The Highlighter Dataset object must have at least the split names
        "train" and "test" and optionally "val". If "val" is not a split
        "test" will be duplicated in the data.yaml

        Args:
            output_dir: Root level directory for the output dataset
            category_attribute_id: Which attribute to use for the detection categories
            categories: Select a subset of the categories. If None will use all
            image_cache_dir: Directory of locally stored images. The yolo
            <train|val|test>/images directories will contain symlinks to these files.
        """
        self.output_dir = Path(output_dir)
        self.labels_dir = self.output_dir / "labels"
        self.images_dir = self.output_dir / "images"
        self.yaml_path = self.output_dir / "data.yaml"
        self.category_attribute_id = category_attribute_id
        self.categories = categories
        self.image_cache_dir = Path(image_cache_dir).absolute()
        self.task: self.TASK = self.TASK(task)

    @staticmethod
    def generate_data_config_dict(
        root_dir: Path,
        train_image_dir: Path,
        val_image_dir: Path,
        test_image_dir: Optional[Path],
        categories: List[str],
    ) -> dict:
        config = {
            "path": str(root_dir),
            "train": str(train_image_dir.relative_to(root_dir)),
            "val": str(val_image_dir.relative_to(root_dir)),
            "names": {i: name for i, name in enumerate(categories)},
        }
        if test_image_dir is not None:
            config["test"] = str(test_image_dir.relative_to(root_dir))

        return config

    def _to_box_label(self, cat_idx, segmentation, scale_wh):
        im_w, im_h = scale_wh

        box_left, box_top, box_w, box_h = segmentation_to_bbox(segmentation)
        box_cen_x = (box_left + box_w / 2) / im_w
        box_cen_y = (box_top + box_h / 2) / im_h
        box_w /= im_w
        box_h /= im_h
        return f"{cat_idx} {box_cen_x} {box_cen_y} {box_w} {box_h}"

    def _to_seg_label(self, cat_idx, segmentation, scale_wh):
        if len(segmentation) > 1:
            warn("YoloWriter only supports " "single polygon, not multipolygon.")
        segmentation_arr = np.array(segmentation[0])
        scale = 1 / np.array(scale_wh * (segmentation_arr.shape[0] // 2))
        scaled_segmentation = (segmentation_arr * scale).tolist()
        segmentation_str = " ".join([str(s) for s in scaled_segmentation])
        return f"{cat_idx} {segmentation_str}"

    def write(
        self,
        dataset: "Dataset",
    ):
        unique_splits = dataset.data_files_df.split.unique()
        if not all([s in unique_splits for s in ("train", "val")]):
            raise ValueError(
                "data_files_df must have at split names "
                "['train', 'val'] optionally 'test'. Got: "
                f"{unique_splits}"
            )
        self.labels_dir.mkdir(parents=True, exist_ok=True)
        self.images_dir.mkdir(parents=True, exist_ok=True)

        unique_categories = dataset.annotations_df[
            dataset.annotations_df.attribute_id == self.category_attribute_id
        ].value.unique()
        unique_categories.sort()
        unique_categories = unique_categories.tolist()

        # Allign categories with passed in categories
        # else fall back on sorted unique_categories
        if self.categories is not None:
            if not all([c in unique_categories for c in self.categories]):
                raise ValueError(
                    "All categories must appear in the source dataset "
                    f"got: categories = {self.categories} and "
                    f"source dataset categories = {unique_categories}"
                )
            categories_of_interest: List[str] = [str(c) for c in self.categories]
        else:
            categories_of_interest: List[str] = [str(c) for c in unique_categories]

        config_dict = self.generate_data_config_dict(
            self.output_dir,
            self.images_dir / "train",
            self.images_dir / "val",
            self.images_dir / "test" if "test" in unique_splits else None,
            categories_of_interest,
        )
        with (self.output_dir / "data.yaml").open("w") as f:
            yaml.dump(config_dict, f)

        def to_yolo_annotation(grp, categories_of_interest=categories_of_interest):
            data = grp.set_index("entity_id").to_dict("index")
            data_file_id = grp.name

            labels = []
            entities = grp.set_index("entity_id").to_dict("index")
            for entity_id, data in entities.items():
                pixel_loc_uuid_str = str(PIXEL_LOCATION_ATTRIBUTE_UUID)
                if pixel_loc_uuid_str not in data:
                    raise ValueError(
                        f"each entity must have a {PIXEL_LOCATION_ATTRIBUTE_UUID} attribute, got: {data} for entity_id {entity_id}"
                    )

                category_uuid_str = str(self.category_attribute_id)
                if category_uuid_str not in data:
                    raise ValueError(
                        f"each entity must have a {category_uuid_str} attribute, got: {data} for entity_id {entity_id}"
                    )

                category = data[category_uuid_str]
                cat_idx = categories_of_interest.index(str(category))

                im_w = data["width"]
                im_h = data["height"]
                pixel_location = data[pixel_loc_uuid_str]
                segmentation = shapely_to_segmentation(pixel_location)

                if self.task == self.TASK.DETECT:
                    labels.append(self._to_box_label(cat_idx, segmentation, (im_w, im_h)))
                elif self.task == self.TASK.SEGMENT:
                    labels.append(self._to_seg_label(cat_idx, segmentation, (im_w, im_h)))

            split_name = grp.split.values[0]
            labels_path = self.labels_dir / split_name / f"{data_file_id}.txt"
            labels_path.parent.mkdir(exist_ok=True)
            with labels_path.open("w") as f:
                f.write("\n".join(labels))

            filename = grp.filename.values[0]
            image_symlink: Path = self.images_dir / split_name / filename
            image_symlink.parent.mkdir(parents=True, exist_ok=True)
            if not image_symlink.exists():
                image_symlink.symlink_to(self.image_cache_dir / filename)

        for split_name in dataset.data_files_df.split.unique():

            ddf = dataset.data_files_df[dataset.data_files_df.split == split_name]
            split_ids = ddf.data_file_id.unique()
            adf = dataset.annotations_df[dataset.annotations_df.data_file_id.isin(split_ids)]
            adf = adf.drop_duplicates(subset=["entity_id", "attribute_id", "value"], keep="first")

            pix_df = adf[adf.attribute_id == str(PIXEL_LOCATION_ATTRIBUTE_UUID)]
            cat_df = adf[
                (adf.attribute_id == str(self.category_attribute_id))
                & (adf.value.isin(categories_of_interest))
            ]

            pix_cat_df = pd.concat([pix_df, cat_df])
            # Add width, height and split columns annotations_df
            df = pd.merge(
                pix_cat_df,
                ddf[["data_file_id", "width", "height", "split", "filename"]],
                on="data_file_id",
                how="left",
            )

            # Pivot the DataFrame to reshape based on 'entity_id' and 'attribute_id'
            pivoted_df = df.pivot(index="entity_id", columns="attribute_id", values="value")

            # Reset the index to make 'id' a column again
            pivoted_df = pivoted_df.reset_index()

            # Fill missing values with None
            pivoted_df = pivoted_df.where(pd.notnull(pivoted_df), None)

            # Drop duplicates for 'width' and 'height' columns, keeping only the first occurrence for each 'id'
            width_height_df = df[
                ["entity_id", "width", "height", "data_file_id", "split", "filename"]
            ].drop_duplicates()

            # Merge the pivoted DataFrame with the width and height DataFrame
            # columns:
            # entity_id | attr_id0 | attr_idN | width | height | data_file_id | split
            merged_df = pd.merge(pivoted_df, width_height_df, on="entity_id")

            merged_df.groupby("data_file_id").apply(to_yolo_annotation)



from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union
from uuid import UUID, uuid4

import aiko_services as aiko
from aiko_services import (
    PROTOCOL_PIPELINE,
    PipelineImpl,
    compose_instance,
    pipeline_args,
    pipeline_element_args,
)
from pydantic import BaseModel
from shapely import geometry as geom

from highlighter.client import EAVT, Annotation, Entity

__all__ = [
    "Annotation",
    "Capability",
    "ContextPipelineElement",
    "Entities",
    "EntityUUID",
    "PROTOCOL_PIPELINE",
    "PipelineElement",
    "PipelineImpl",
    "StreamEvent",
    "compose_instance",
    "compose_instance",
    "pipeline_args",
    "pipeline_element_args",
]

EntityUUID = UUID


class Entities(dict):

    def add_entity(
        self,
        entity_id: Optional[UUID] = None,
        location: Optional[geom.Polygon] = None,
        observations: List[EAVT] = [],
        global_observations: List[EAVT] = [],
    ):
        if entity_id is None:
            entity_id = uuid4()

        annotations = []
        if observations:
            annotations = [
                Annotation(
                    id=uuid4(),
                    entity_id=entity_id,
                    location=location,
                    observations=observations,
                    datum_source=observations[0].datum_source,
                )
            ]

        entity = Entity(
            id=entity_id,
            annotations=annotations,
            global_observations=global_observations,
        )

        super().__setitem__(entity_id, entity)


"""Decouple the rest of the code from aiko.PipelineElement"""
ContextPipelineElement = aiko.ContextPipelineElement
StreamEvent = aiko.StreamEvent
PipelineElement = aiko.PipelineElement


class Capability(PipelineElement):

    class DefaultStreamParameters(BaseModel):
        """Populate with default stream param key fields"""

        pass

    def __init__(self, context: aiko.ContextPipelineElement):
        context.get_implementation("PipelineElement").__init__(self, context)

    @classmethod
    def default_stream_parameters(cls) -> BaseModel:
        return cls.DefaultStreamParameters()

    @classmethod
    def get_capability_input_output_deploy(cls) -> Dict:
        return {}

    def _get_parameter(self, name, default=None, use_pipeline=True) -> Tuple[Any, bool]:
        """Adds the correct output type to get_parameter type checking
        does not complain
        """
        return self.get_parameter(name, default, use_pipeline)

    @classmethod
    def get_capability_definition(cls) -> Dict:
        input_output_deploy = cls.get_capability_input_output_deploy()
        capability_definition = dict(
            name=cls.__name__,
            **input_output_deploy,
        )

        default_parameters = cls.default_stream_parameters().model_dump()
        capability_definition["parameters"] = default_parameters
        return capability_definition

    def start_stream(self, stream, stream_id) -> Tuple[StreamEvent, Optional[str]]:
        """Called when a stream is created"""
        # If there are name clashes between capabiliteis then
        # the parameter that is in the capability later in the pipeline will take
        # precedence. To get around this you can prepend the stream parameter
        # name with the Capability name in the cli
        # "-sp CapabilityA.foo=42 -sp CapabilityB.foo=3.14
        return StreamEvent.OKAY, None

    def stop_stream(self, stream, stream_id) -> Tuple[StreamEvent, Optional[str]]:
        """Called when a stream is terminated"""
        return StreamEvent.OKAY, None

    def process_frame(self, stream, image: Union[str, Path, bytes]) -> Tuple[StreamEvent, Union[Dict, str]]:
        """Called once per frame"""
        raise NotImplementedError()
